
/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



vec3 calculateAtmosphere(in vec3 rayDir, in mat2x3 sc, in mat3x3 ec, in vec3 illuminanceMulti, const int steps) {
    vec3 viewPos = vec3(0.0, planetRadius + eyeAltitude, 0.0);

    vec2 atmosphereDists = rsi(viewPos, rayDir, atmosphereRadius);
    vec2 planetDists = rsi(viewPos, rayDir, atmosphereLowerLimit);

	bool planetIntersected = planetDists.y >= 0.0;
   	bool atmosphereIntersected = atmosphereDists.y >= 0.0;

    vec2 sd = vec2(0.0);

    if (planetIntersected && planetDists.x < 0.0) sd.x  = planetDists.y;
    else sd.x  = max(atmosphereDists.x, 0.0);

    if (planetIntersected && planetDists.x > 0.0) sd.y  = planetDists.x;
    else sd.y  = atmosphereDists.y;

	float stepSize = length(sd.y - sd.x) * rcp(steps);
	vec3  increment = rayDir * stepSize;
	vec3 position = rayDir * sd.x + (increment * 0.3 + viewPos);

    vec3 scatteredSun = vec3(0.0);
    vec3 scatteredMoon = vec3(0.0);
    vec3 scatteredMulti = vec3(0.0);
    vec3 transmittance = vec3(1.0);

    vec4 phase = vec4(phaseFunction(dot(rayDir, sunVector)), phaseFunction(dot(rayDir, moonVector)));
    float isotropicPhase = 0.25 * pi;

    for(int i = 0; i < steps; ++i, position += increment) {
        vec3 airmass = calculateAtmosphereDensity(length(position));
        if(airmass.y > 1e35) break;
        vec3 opticalDepth = stepSize * airmass;
        vec3 stepOpticalDepth = ec * opticalDepth;

        vec3 stepTransmittance = saturate(expf(-stepOpticalDepth));
        vec3 stepTransmittedFraction = saturate((stepTransmittance - 1.0) / -stepOpticalDepth);
        vec3 visibleScattering = transmittance * stepTransmittedFraction;

        mat2x3 stepScatteringView = sc * mat2x2(opticalDepth.xy * phase.xy, opticalDepth.xy * phase.zw);
        stepScatteringView[0] *= visibleScattering;
        stepScatteringView[1] *= visibleScattering;

        scatteredSun += stepScatteringView[0] * calculateAtmosphereTransmittance(sunVector, position);
        scatteredMoon += stepScatteringView[1] * calculateAtmosphereTransmittance(moonVector, position, ec, 8);     //night is so dark that the extra quality is negligible

        scatteredMulti += sc * (opticalDepth.xy * isotropicPhase) * visibleScattering;

        transmittance *= stepTransmittance;
    }

    vec3 scattering = (scatteredSun * sunIlluminance) + (scatteredMoon * moonIlluminance) + (scatteredMulti * illuminanceMulti);

    return max(scattering, 0.0);
}

vec3 calculateAtmosphere(in vec3 rayDir, in vec3 illuminanceMulti) {
    return calculateAtmosphere(rayDir, scatteringCoefficients, extinctionCoefficients, illuminanceMulti, SCATTERING_STEPS);
}

#include "spectralCoeffs.glsl"

vec3 calculateSpectralAtmosphere(in vec3 rayDir) {
    vec3 xyz = vec3(0.0);

    int wlCount;
    for(int i = 390; i < 830; i += 50) {
		vec3 wl = vec3(i);
        mat2x3 scattering = mat2x3(betaR(wl)*atmos_rayCoeff, betaM());
        mat3x3 extinction = mat3x3(betaR(wl)*atmos_rayCoeff, betaM() * 1.11, spectralOzone(wl));

		vec3 atmosphereUp = calculateAtmosphere(vec3(0, 1, 0), scattering, extinction, vec3(0.0), SCATTERING_STEPS/2);
		vec3 atmosphere = calculateAtmosphere(rayDir, scattering, extinction, atmosphereUp, SCATTERING_STEPS);

        xyz += atmosphere.x * spectrumToXYZ(wl.x);
        xyz += atmosphere.y * spectrumToXYZ(wl.y);
        xyz += atmosphere.z * spectrumToXYZ(wl.z);


        wlCount++;
    }

    vec3 rgb = xyzToRGB(xyz);

    return rgb / wlCount;
}
