/*
drop in replacement for an updated version of the raymarcher for clouds
computes both volumetric and planar clouds in a single function to save redundant operations
*/

vec4 computeCloudLayersReflect(vec3 worldDir, float dither, vec3 atmosColor) {
    const float eyeAltitude = 64.0;
    vec3 camPos     = vec3(cameraPosition.x, eyeAltitude, cameraPosition.z);
    vec2 planetSphere   = rsi(vec3(0.0, planetRadius + eyeAltitude, 0.0), worldDir, planetRadius);

    bool occluded       = (eyeAltitude < cloudVolumeMinY && planetSphere.y > 0.0) || (eyeAltitude > clayer1_maxY && worldDir.y > 0.0);
    if (occluded) return vec4(0.0, 0.0, 0.0, 1.0);

    /* ------ universal variables and values ------ */
    const float sigmaA  = 1.00;         //absorption coefficient, not physically based
    const float sigmaS  = 0.10;         //scattering coefficient, can be assumed to be sigmaT if the albedo is close to 1.0
    const float sigmaT  = 0.10;         //extinction coefficient, 0.05-0.12 for cumulus, 0.04-0.06 for stratus

    float totalTransmittance = 1.0;
    float globalTransmittance = 1.0;
    float cosTheta      = dot(worldDir, cloudLightvector);
    float powderIntensity = saturate(cornettShanksMiePhase(cosTheta, 0.5));

    vec3 sunlight       = lightColor[0];
    vec3 skylight       = lightColor[1] * 0.25;
    vec3 bouncelight    = (lightColor[0] + lightColor[1]) * (0.01 * vec3(0.7, 1.0, 0.6));

    vec3 atmosAirmass   = calculateAtmosphereDensity(length(worldDir) * worldDir.y * 0.35 + eyeAltitude + planetRadius);
    vec3 atmosStepOD    = extinctionCoefficients * atmosAirmass;

    vec3 totalColor     = vec3(0.0);
    vec3 rpos           = vec3(0.0);

    #if (defined cloudLayer0Mode || defined cloudLayer1Mode)

    /* ------ raymarched clouds ------ */
    vec2 bSphere        = rsi(vec3(0.0, planetRadius+eyeAltitude, 0.0), worldDir, planetRadius+cloudVolumeMinY);
    vec2 tSphere        = rsi(vec3(0.0, planetRadius+eyeAltitude, 0.0), worldDir, planetRadius+clayer1_maxY);

    float startDist     = eyeAltitude>clayer1_maxY ? tSphere.x : bSphere.y;
    float endDist       = eyeAltitude>clayer1_maxY ? bSphere.x : tSphere.y;

    vec3 startPos       = worldDir * startDist;
        startPos        = planetCurvePosition(startPos);
    vec3 endPos         = worldDir * endDist;
        endPos          = planetCurvePosition(endPos);

    const float baseStep = 70 * vcloud_quality;
    const float baseLength = cloudVolumeTotalDepth * rcp(baseStep);

    float stepLength    = distance(endPos, startPos) * rcp(baseStep);
    float stepCoeff     = stepLength * rcp(baseLength);
        stepCoeff       = 0.45 + clamp(stepCoeff - 1.2, 0.0, 3.5) * 0.5;

    uint steps          = uint(baseStep * stepCoeff);

    vec3 rstep          = (endPos - startPos) * rcp(float(steps));
        rpos            = rstep * dither + startPos + camPos;
    float rlength       = length(rstep);

    float fadeDistance  = -1.0;

    vec3 powderParam    = vec3(30.0, 2.0, 1.0);

    for (uint i = 0; i < steps; ++i, rpos += rstep) {
        if (globalTransmittance<vcloud_thresh) break;
        if (rpos.y<cloudVolumeMinY || rpos.y>clayer1_maxY) continue;

        float dist      = distance(rpos, camPos);
        if (dist>=vcloud_clip) continue;
        float dfade     = saturate(dist/vcloud_clip);

        float density   = volumeCloudContainer(rpos, powderParam);
        if (density<=0.0) continue;

        float extinction = density * sigmaT;
        float stept     = expf(-extinction*rlength);
        float integral  = (1.0 - 1.0*stept) / sigmaT;           

        fadeDistance    = max(fadeDistance, 0.0) + dist * globalTransmittance * integral * sigmaT;

        float directod  = volumeCloudDirectOD(rpos, vcloud_direct_steps)*sigmaA;
        float skyod     = volumeCloudSkyOD(rpos, vcloud_skylight_steps)*sigmaA;

        float p         = sstep(rpos.y, 3e3, 12e3);

        float powder    = 1.0 - expf(-density * powderParam.x) * powderParam.z;
        float dpowder   = min(powder, (1.0-expf(-directod * powderParam.y) * powderParam.z) * sqrt(powder));
            dpowder     = mix(dpowder, 1.0, powderIntensity);

        vec3 scatter = vec3(0.0);

        float s_d       = sigmaS * 2.0;      //scatter derivate
        float t_d       = sigmaT * 2.0;      //transmittance/attentuation derivate

        for (uint j = 0; j < 7u; ++j) {
            s_d        *= 0.5;
            t_d        *= 0.5;
            
            float phase = cloud_phase(cosTheta, pow(0.5, float(j)));  //phase derivate

            scatter.x  += expf(-directod*t_d) * phase * s_d;
            scatter.y  += expf(-skyod*t_d) * s_d;
            scatter.z  += volumeCloudBouncedLight(rpos) * s_d;
        }

        vec3 resultScatter = sunlight * (scatter.x * dpowder) + skylight * (scatter.y * powder) + bouncelight * (scatter.z  * (powder * 0.5 + 0.5));
        
        vec3 atmosTransmittance = expf(-atmosStepOD * dist) * expf(-dist * 6e-6);
        resultScatter   = mix(atmosColor * sigmaT, resultScatter, atmosTransmittance);

        totalColor     += resultScatter * (integral * globalTransmittance);

        globalTransmittance  *= stept;
    }

    if (fadeDistance < -0.5) fadeDistance = vcloud_clip;

    globalTransmittance = linStep(globalTransmittance, vcloud_thresh, 1.0);

    float fade          = linStep(fadeDistance, 0.75 * vcloud_clip, 0.99 * vcloud_clip);
        fade            = saturate(1.0-fade);
        fade            = sqr(fade);

    totalColor         *= fade;
    totalTransmittance  = mix(1.0, globalTransmittance, fade);

    #endif

    /* ------ planar clouds ------ */
    #if vcloud_highlayer != 0
    if (globalTransmittance > 1e-8) {     //planar cirrus layer
        vec2 cirrSphere = rsi(vec3(0.0, planetRadius + eyeAltitude, 0.0), worldDir, planetRadius + cloudLayer2PlaneY);

        float posDist   = eyeAltitude > cloudLayer2PlaneY ? cirrSphere.x : cirrSphere.y;

        rpos            = camPos + worldDir * posDist + worldDir * (dither) * (cloudLayer2PlanarMaxY - cloudLayer2PlanarMinY) * 0.08;
        const float rlength = cloudLayer2PlaneY * 0.8;

        float dist      = distance(rpos, camPos);

        if (dist < vcloud_clip) {
            float dfade     = saturate(dist/vcloud_clip);

            float powder_k  = 30.0;
            float density   = cloudLayer2(rpos, rpos.y);

            if (density > 0.0) {
                float extinction = density * sigmaT;
                float stept     = expf(-extinction * rlength);
                float integral  = (1.0 - 1.0*stept) / sigmaT;

                float directod  = planarCloudL2DirectOD(rpos, 4) * sigmaA;
                float skyod     = planarCloudL2SkyOD(rpos, 2) * sigmaA;

                float powder    = 1.0 - expf(-density * powder_k) * 0.7;
                float dpowder   = mix(powder, 1.0, powderIntensity);

                vec2 scatter    = vec2(0.0);

                float s_d       = sigmaS * 2.0;      //scatter derivate
                float t_d       = sigmaT * 2.0;      //transmittance/attentuation derivate

                for (uint j = 0; j < 7u; ++j) {
                    s_d        *= 0.5;
                    t_d        *= 0.5;
                    
                    float phase = cloud_phase(cosTheta, pow(0.5, float(j)));  //phase derivate

                    scatter.x  += expf(-directod*t_d) * phase * s_d;
                    scatter.y  += expf(-skyod*t_d) * s_d;
                }

                vec3 resultScatter = sunlight * (scatter.x * dpowder) + skylight * (scatter.y * powder);
                
                vec3 atmosTransmittance = expf(-atmosStepOD * dist) * expf(-dist * 6e-6);
                resultScatter   = mix(atmosColor * sigmaT, resultScatter, atmosTransmittance);
            
                float fade      = linStep(dist, 0.75 * vcloud_clip, 0.99 * vcloud_clip);
                    fade        = saturate(1.0-fade);
                    fade        = sqr(fade);

                totalColor     += resultScatter * (integral * globalTransmittance * fade);
                globalTransmittance *= stept;
                totalTransmittance *= mix(1.0, stept, fade);
            }
        }
    }
    #endif

    #ifdef cloudLayer3Mode
    if (globalTransmittance > 1e-8) {     //planar cirrus layer 2
        vec2 cirrSphere = rsi(vec3(0.0, planetRadius + eyeAltitude, 0.0), worldDir, planetRadius + cloudLayer3PlaneY);

        float posDist   = eyeAltitude > cloudLayer3PlaneY ? cirrSphere.x : cirrSphere.y;

        rpos            = camPos + worldDir * posDist + worldDir * (dither) * (cloudLayer3PlaneDepth * 0.08);
        const float rlength = cloudLayer3PlaneY * 0.8;

        float dist      = distance(rpos, camPos);

        if (dist < vcloud_clip) {
            float dfade     = saturate(dist/vcloud_clip);

            float powder_k  = 30.0;
            float density   = cloudLayer3(rpos, rpos.y);

            if (density > 0.0) {
                float extinction = density * sigmaT;
                float stept     = expf(-extinction * rlength);
                float integral  = (1.0 - 1.0*stept) / sigmaT;

                float directod  = planarCloudL3DirectOD(rpos, 3) * sigmaA;
                float skyod     = planarCloudL3SkyOD(rpos, 2) * sigmaA;

                float powder    = 1.0 - expf(-density * powder_k) * 0.7;
                float dpowder   = mix(powder, 1.0, powderIntensity);

                vec2 scatter    = vec2(0.0);

                float s_d       = sigmaS * 2.0;      //scatter derivate
                float t_d       = sigmaT * 2.0;      //transmittance/attentuation derivate

                for (uint j = 0; j < 7u; ++j) {
                    s_d        *= 0.5;
                    t_d        *= 0.5;
                    
                    float phase = cloud_phase(cosTheta, pow(0.5, float(j)));  //phase derivate

                    scatter.x  += expf(-directod*t_d) * phase * s_d;
                    scatter.y  += expf(-skyod*t_d) * s_d;
                }

                vec3 resultScatter = sunlight * (scatter.x * dpowder) + skylight * (scatter.y * powder);
                
                vec3 atmosTransmittance = expf(-atmosStepOD * dist) * expf(-dist * 6e-6);
                resultScatter   = mix(atmosColor * sigmaT, resultScatter, atmosTransmittance);
            
                float fade      = linStep(dist, 0.75 * vcloud_clip, 0.99 * vcloud_clip);
                    fade        = saturate(1.0-fade);
                    fade        = sqr(fade);

                totalColor     += resultScatter * (integral * globalTransmittance * fade);
                globalTransmittance *= stept;
                totalTransmittance *= mix(1.0, stept, fade);
            }
        }
    }
    #endif

    return vec4(totalColor, totalTransmittance);
}