
/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/


#ifdef g_textured

    #if texture_format==13
        int lab_f0(in float f0) {
            return int(round(f0 * 255.0));
        }
    #elif texture_format==12 || texture_format==11
        int lab_f0(in float f0) {
            return int(round(sqrt(f0) * 255.0));
        }
    #endif

    #if defined specular_resourcepack && defined normals_resourcepack
        #if texture_format==13 || texture_format==12
            vec3 decode_lab_nrm(in vec3 ntex, in mat3x3 tbn, inout float ao) {
                vec3 nrm = ntex * 2.0 - (254.0 * rcp(255.0));

                if(any(greaterThan(ntex, vec3(0.003)))) ao *= ntex.z;
                nrm.z    = sqrt(saturate(1.0 - dot(nrm.xy, nrm.xy)));

                return normalize(nrm * tbn);
            }
        #elif texture_format==11
            vec3 decode_lab_nrm(in vec3 ntex, in mat3x3 tbn, inout float ao) {
                vec3 nrm = ntex * 2.0 - (254.0 * rcp(255.0));

                if(any(greaterThan(ntex, vec3(0.003)))) ao *= sqr(length(nrm));

                return normalize(nrm * tbn);
            }
        #endif
    #elif defined normals_resourcepack
        vec3 decode_nrm(in vec3 ntex, in mat3x3 tbn) {
            vec3 nrm = ntex * 2.0 - (254.0 * rcp(255.0));

            nrm.z    = sqrt(saturate(1.0 - dot(nrm.xy, nrm.xy)));

            return normalize(nrm * tbn);
        }
    #endif

#else

    struct pbr_data {
        float roughness;
        float roughnessSquared;
        float f0;
        float porosity;
        float sss;
        float emissive;

        bool is_metal;
        bool generic_metal;

        vec3 eta;
        vec3 etak;
    };

    float f0ToIOR(float f0) {
        f0 = sqrt(f0);
        f0 *= 0.99999; // Prevents divide by 0
        return (1.0 + f0) / (1.0 - f0);
    }

    float IORTof0(in float ior) {
        float sqrtf0 = (ior - 1.00029) / (ior + 1.00029);
        return sqr(sqrtf0);
    }

    vec3 fresnelConductor(in vec3 eta, in vec3 etak, in float cosTheta) {  
        /*
            Should be the same function as the dielectrics, but that function only works for dielectrics right now, so we are doing this instead.
        */
        float cosTheta2 = cosTheta * cosTheta;
        float sinTheta2 = 1 - cosTheta2;
        vec3 eta2 = eta * eta;
        vec3 etak2 = etak * etak;

        vec3 t0 = eta2 - etak2 - sinTheta2;
        vec3 a2plusb2 = sqrt(t0 * t0 + 4 * eta2 * etak2);
        vec3 t1 = a2plusb2 + cosTheta2;
        vec3 a = sqrt(0.5f * (a2plusb2 + t0));
        vec3 t2 = 2 * a * cosTheta;
        vec3 Rs = (t1 - t2) / (t1 + t2);

        vec3 t3 = cosTheta2 * a2plusb2 + sinTheta2 * sinTheta2;
        vec3 t4 = t2 * sinTheta2;   
        vec3 Rp = Rs * (t3 - t4) / (t3 + t4);

        return saturate(0.5 * (Rp + Rs));
    }

    vec3 fresnelNonPolarized(in float voh, in vec3 n2) {
        /*
            Equation from https://en.wikipedia.org/wiki/Fresnel_equations
            Currently only works for dielectrics.
        */
        const vec3 n1 = vec3(1.00029);
        vec3 cosThetaI = vec3(voh);
        float sinThetaI = sqrt(saturate(1.0 - voh * voh));
        vec3 sinThetaT = sinThetaI * n1 / n2;
        vec3 cosThetaT = sqrt(1 - sqr(sinThetaT));

        vec3 Rs = sqr(
            ((n1 * cosThetaI) - (n2 * cosThetaT)) / 
            ((n1 * cosThetaI) + (n2 * cosThetaT))
        );

        vec3 Rp = sqr(
            ((n1 * cosThetaT) - (n2 * cosThetaI)) / 
            ((n1 * cosThetaT) + (n2 * cosThetaI))
        );

        return saturate((Rs + Rp) * 0.5);
    }

    vec3 get_fresnel(in vec3 eta, in vec3 etak, in float cosTheta, bool is_metal) {
        if(is_metal) {
            return fresnelConductor(eta, etak, cosTheta);
        } else {
            return fresnelNonPolarized(cosTheta, eta);
        }
    }

    void lab_metals(inout pbr_data pbr, in float f0_index) {
        if(f0_index == 230) {         //iron
            pbr.eta  = vec3(2.9114, 2.9497, 2.5845);
            pbr.etak = vec3(3.0893, 2.9318, 2.7670);
        } else if(f0_index == 231) {  //gold
            pbr.eta  = vec3(0.18299, 0.42108, 1.3734);
            pbr.etak = vec3(3.4242, 2.3459, 1.7704);
        } else if(f0_index == 232) {  //aluminum
            pbr.eta  = vec3(1.3456, 0.96521, 0.61722);
            pbr.etak = vec3(7.4746, 6.3995, 5.3031);
        } else if(f0_index == 233) {  //chrome
            pbr.eta  = vec3(3.1071, 3.1812, 2.3230);
            pbr.etak = vec3(3.3314, 3.3291, 3.1350);
        } else if(f0_index == 234) {  //copper
            pbr.eta  = vec3(0.27105, 0.67693, 1.3164);
            pbr.etak = vec3(3.6092, 2.6248, 2.2921);
        } else if(f0_index == 235) {  //lead
            pbr.eta  = vec3(1.9100, 1.8300, 1.4400);
            pbr.etak = vec3(3.5100, 3.4000, 3.1800);
        } else if(f0_index == 236) {  //platinum
            pbr.eta  = vec3(2.3757, 2.0847, 1.8453);
            pbr.etak = vec3(4.2655, 3.7153, 3.1365);
        } else if(f0_index == 237) {  //silver
            pbr.eta  = vec3(0.15943, 0.14512, 0.13547);
            pbr.etak = vec3(3.9291, 3.1900, 2.3808);
        }
    }

    #if texture_format==13
        pbr_data decode_lab(vec4 unpacked_tex) {
            pbr_data pbr;

            pbr.roughness   = (1.0-unpacked_tex.x);
            pbr.roughnessSquared = sqr(pbr.roughness);
            pbr.f0          = unpacked_tex.y;
            pbr.porosity    = unpacked_tex.z<(64.5/255.0) ? linStep(unpacked_tex.z*255.0, 0.0, 64.0) : 0.0;
            pbr.sss         = linStep(unpacked_tex.z*255.0, 64.0, 255.0);
            pbr.emissive    = linStep(unpacked_tex.a*255.0, 0.0, 254.0);

            pbr.eta = vec3(f0ToIOR(pbr.f0));
            pbr.etak = vec3(0.0);

            int f0_index    = int(unpacked_tex.y*255.0);

            pbr.is_metal    = f0_index>229;
            pbr.generic_metal = f0_index>254;

            if(pbr.is_metal) lab_metals(pbr, f0_index);
            else if (pbr.generic_metal) pbr.eta = vec3(f0ToIOR(0.25));

            return pbr;
        }
    #elif texture_format==12 || texture_format==11
        pbr_data decode_lab(vec4 unpacked_tex) {
            pbr_data pbr;

            pbr.roughness   = (1.0-unpacked_tex.x);
            pbr.roughnessSquared = sqr(pbr.roughness);
            pbr.f0          = sqr(unpacked_tex.y);
            pbr.porosity    = unpacked_tex.z<(64.5/255.0) ? linStep(unpacked_tex.z*255.0, 0.0, 64.0) : 0.0;
            pbr.sss         = linStep(unpacked_tex.z*255.0, 64.0, 255.0);
            pbr.emissive    = linStep(unpacked_tex.a*255.0, 0.0, 254.0);

            pbr.eta = vec3(f0ToIOR(pbr.f0));
            pbr.etak = vec3(0.0);

            int f0_index    = int(unpacked_tex.y*255.0);

            pbr.is_metal    = f0_index>229;
            pbr.generic_metal = f0_index>254;

            if(pbr.is_metal) lab_metals(pbr, f0_index);

            return pbr;
        }
    #endif

#endif