
/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"

#if (ibl_mode != 1 && defined IBL_pass)
    #undef IBL_pass
#endif


/* ------ outputs ------ */
out vec2 coord;

#ifdef ONLY_BLOCKLIGHT
    flat out float lightFlip;

    flat out vec3 blockLightColor;
#else
    flat out float lightFlip;

    flat out mat3x3 lightColor;

    #ifdef IBL_pass
        flat out mat4x3 IBL_light;
    #endif
#endif

#ifdef STATIC_LIGHTVECTOR
    out vec3 lightVector;
#endif

/* ------ uniforms ------ */
uniform float eyeAltitude;
uniform float wetness;

uniform vec2 viewSize;

uniform vec3 sunVector, sunVectorView;
uniform vec3 moonVector, moonVectorView;
uniform vec3 sunPosition, upPosition;

uniform sampler2D colortex4;

/* ------ includes ------ */
#include "/lib/util/srgb.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/blackbody.glsl"
#include "/lib/atmosphere/constants.glsl"
#include "/lib/atmosphere/transmittance.glsl"
#include "/lib/atmosphere/project.glsl"

vec2 unproject_sphere(vec3 dir) {
    vec2 lonlat     = vec2(atan(-dir.x, dir.z), acos(dir.y));
    return lonlat * vec2(rcp(tau), rcp(pi)) + vec2(0.5, 0.0);
}
vec3 get_projected_skycloud(vec3 x) {
    vec2 spherecoord = unproject_sphere(x);
    spherecoord = spherecoord*rcp(SKYREF_LOD) + vec2(0.5, 0.0);

    return texture(colortex4, spherecoord).rgb;
}

vec3 get_projected_sky(vec3 x) {
    return texture(colortex4, projectSky(x)).rgb;
    //return get_projected_skycloud(x);
}

#ifdef STATIC_LIGHTVECTOR
vec2 rotate_coord(vec2 pos, const float angle) {
    return vec2(cos(angle)*pos.x + sin(angle)*pos.y, 
                cos(angle)*pos.y - sin(angle)*pos.x);
}
#endif

void main() {
    gl_Position         = vec4(gl_Vertex.xy * 2.0 - 1.0, 0.0, 1.0);
    coord               = gl_MultiTexCoord0.xy;

    #ifdef ONLY_BLOCKLIGHT
        blockLightColor = blackbody(blocklightTemp) * lux(2000) * blocklightMult;
    #else
        lightColor[0]   = dot(sunPosition, upPosition) > 0.5 ? calculateAtmosphereTransmittance(sunVector, vec3(0.0, planetRadius + eyeAltitude, 0.0)) * sunIlluminance : calculateAtmosphereTransmittance(moonVector, vec3(0.0, planetRadius + eyeAltitude, 0.0)) * moonIlluminance;
        lightColor[1]   = get_projected_sky(vec3(0.0, 1.0, 0.0)) * pi;
        lightColor[1]   = mix(lightColor[1], vec3(v3avg(lightColor[1]) * 1.5), wetness * 0.6);
        lightColor[2]   = blackbody(blocklightTemp) * lux(2000) * blocklightMult;

        float lf        = dot(sunVector, vec3(0.0, 1.0, 0.0))*0.5+0.5;
            lf          = linStep(lf, 0.48, 0.499)*(1.0-linStep(lf, 0.501, 0.52));

        lightFlip       = saturate(1.0-lf);
    #endif

    #ifdef STATIC_LIGHTVECTOR
        lightVector    = vec3(0.0, 1.0, 0.0);
        lightVector.yz = rotate_coord(lightVector.yz, radians(sunPathRotation));
        lightVector    = normalize(lightVector);

        #ifdef ONLY_BLOCKLIGHT
            float lf        = dot(lightVector, vec3(0.0, 1.0, 0.0))*0.5+0.5;
                lf          = linStep(lf, 0.48, 0.499)*(1.0-linStep(lf, 0.501, 0.52));

            lightFlip       = saturate(1.0-lf);
        #endif
    #endif

    #ifdef IBL_pass
        mat2x3 IBL_e_1;
        mat2x3 IBL_e_2;
        mat2x3 IBL_e_3;

        IBL_e_1[0]  = get_projected_sky(normalize(vec3(1.0, 0.05, -0.5)));
        IBL_e_1[1]  = get_projected_sky(normalize(vec3(1.0, 0.50, -0.5)));
        IBL_e_2[0]  = get_projected_sky(normalize(vec3(1.0, 0.15,  0.0)));
        IBL_e_2[1]  = get_projected_sky(normalize(vec3(1.0, 0.50,  0.0)));
        IBL_e_3[0]  = get_projected_sky(normalize(vec3(1.0, 0.05,  0.5)));
        IBL_e_3[1]  = get_projected_sky(normalize(vec3(1.0, 0.50,  0.5)));

        IBL_light[0] = IBL_e_1[0] + IBL_e_1[1] + IBL_e_2[0] + IBL_e_2[1] + IBL_e_3[0] + IBL_e_3[1];

        mat2x3 IBL_s_1;
        mat2x3 IBL_s_2;
        mat2x3 IBL_s_3;

        IBL_s_1[0]  = get_projected_sky(normalize(vec3(-0.5, 0.15,  1.0)));
        IBL_s_1[1]  = get_projected_sky(normalize(vec3(-0.5, 0.50,  1.0)));
        IBL_s_2[0]  = get_projected_sky(normalize(vec3( 0.0, 0.15,  1.0)));
        IBL_s_2[1]  = get_projected_sky(normalize(vec3( 0.0, 0.50,  1.0)));
        IBL_s_3[0]  = get_projected_sky(normalize(vec3( 0.5, 0.15,  1.0)));
        IBL_s_3[1]  = get_projected_sky(normalize(vec3( 0.5, 0.50,  1.0)));

        IBL_light[1] = IBL_s_1[0] + IBL_s_1[1] + IBL_s_2[0] + IBL_s_2[1] + IBL_s_3[0] + IBL_s_3[1];

        mat2x3 IBL_w_1;
        mat2x3 IBL_w_2;
        mat2x3 IBL_w_3;

        IBL_w_1[0]  = get_projected_sky(normalize(vec3(-1.0, 0.15, -0.5)));
        IBL_w_1[1]  = get_projected_sky(normalize(vec3(-1.0, 0.50, -0.5)));
        IBL_w_2[0]  = get_projected_sky(normalize(vec3(-1.0, 0.15,  0.0)));
        IBL_w_2[1]  = get_projected_sky(normalize(vec3(-1.0, 0.50,  0.0)));
        IBL_w_3[0]  = get_projected_sky(normalize(vec3(-1.0, 0.15,  0.5)));
        IBL_w_3[1]  = get_projected_sky(normalize(vec3(-1.0, 0.50,  0.5)));

        IBL_light[2] = IBL_w_1[0] + IBL_w_1[1] + IBL_w_2[0] + IBL_w_2[1] + IBL_w_3[0] + IBL_w_3[1];
    
        mat2x3 IBL_n_1;
        mat2x3 IBL_n_2;
        mat2x3 IBL_n_3;

        IBL_n_1[0]  = get_projected_sky(normalize(vec3(-0.5, 0.15, -1.0)));
        IBL_n_1[1]  = get_projected_sky(normalize(vec3(-0.5, 0.50, -1.0)));
        IBL_n_2[0]  = get_projected_sky(normalize(vec3( 0.0, 0.15, -1.0)));
        IBL_n_2[1]  = get_projected_sky(normalize(vec3( 0.0, 0.50, -1.0)));
        IBL_n_3[0]  = get_projected_sky(normalize(vec3( 0.5, 0.15, -1.0)));
        IBL_n_3[1]  = get_projected_sky(normalize(vec3( 0.5, 0.50, -1.0)));

        IBL_light[3] = IBL_n_1[0] + IBL_n_1[1] + IBL_n_2[0] + IBL_n_2[1] + IBL_n_3[0] + IBL_n_3[1];

        IBL_light *= rcp(6.0) * pi;

    #endif
}