#version 450 compatibility

/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:034*/
layout(location = 0) out vec4 sceneColor;
layout(location = 1) out vec4 fogScattering;
layout(location = 2) out vec4 fogTransmittance;


/* ------ inputs from vertex stage ------ */
in vec2 coord;


/* ------ uniforms ------ */
uniform int frameCounter;
uniform int isEyeInWater;

uniform vec2 pixelSize;
uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;


/* ------ includes ------*/
#include "/lib/util/encoders.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/atmosphere/constants.glsl"



/* ------ functions ------ */
vec3 decode3x8(float a){
    int bf = int(a*65535.);
    return vec3(bf%32, (bf>>5)%64, bf>>11) / vec3(31,63,31);
}

/* ------ water volume ------ */
vec3 compute_waterVolume(vec3 scenePos0, vec3 scenePos1, float dither) {
    bool isUnderwater   = isEyeInWater == 1;

    vec3 epos   = isUnderwater ? scenePos0 : scenePos1;
    vec3 spos   = isUnderwater ? gbufferModelViewInverse[3].xyz : scenePos0;

    float bstep      = length((epos-spos));
    float stepcoeff  = saturate(bstep * rcp(64.0));
    uint steps       = 4 + uint(stepcoeff * 6);

    vec3 rstep      = (epos-spos)/steps;
    vec3 rpos       = rstep*dither+spos;
    float rlength   = length(rstep);

    vec3 transmittance = vec3(1.0);

    const vec3 absorbCoeff = vec3(2e-1, 4e-2, 1e-2);
    const vec3 scatterCoeff = vec3(1.5e-2);
    const vec3 attenCoeff = (absorbCoeff + scatterCoeff) * 4.0;

    for (uint i = 0; i < steps; i++, rpos += rstep) {
        if (max(transmittance.x, max(transmittance.y, transmittance.z)) < 0.01) break;

        float od    = rlength;
        transmittance  *= expf(-attenCoeff * od);
    }

    return transmittance;
}

vec3 blendTranslucencies(vec3 sceneCol, vec4 translucents, vec3 albedo) {
    vec3 color  = sceneCol;
        color  *= mix(vec3(1.0), albedo, translucents.a);
        color   = mix(color, translucents.rgb, translucents.a);

    return color;
}

void main() {
    vec4 sceneCol      = stex(colortex0);
    float sceneDepth0  = stex(depthtex0).r;
    float sceneDepth1  = stex(depthtex1).r;

    vec4 tex2          = stex(colortex2);
    int matID          = decodeMatID16(tex2.x);

    float bluenoise    = dither_bluenoise();
    

    vec3 viewPos0      = screenSpaceToViewSpace(vec3(coord, sceneDepth0), gbufferProjectionInverse, true);
    vec3 scenePos0     = viewSpaceToSceneSpace(viewPos0, gbufferModelViewInverse);

    #ifdef vfog_enabled
        vec3 viewPos1  = screenSpaceToViewSpace(vec3(coord, sceneDepth1), gbufferProjectionInverse, true);
        vec3 scenePos1 = viewSpaceToSceneSpace(viewPos1, gbufferModelViewInverse);

        vec3 waterVolumeFogTransmittance;

        if (sceneDepth0 < sceneDepth1) {
            vec4 tex1                   = stex(colortex1);
            vec4 translucentColor       = stex(colortex3);
            vec3 translucentAlbedo      = sqr(decode3x8(tex1.b));

            if (matID == 101 && isEyeInWater == 0) {
                waterVolumeFogTransmittance = compute_waterVolume(scenePos0, scenePos1, bluenoise);
                sceneCol.rgb            = sceneCol.rgb * waterVolumeFogTransmittance;
            }

            sceneCol.rgb                = blendTranslucencies(sceneCol.rgb, translucentColor, translucentAlbedo);
        } 
        
        if (isEyeInWater == 1) {
            waterVolumeFogTransmittance = compute_waterVolume(scenePos0, scenePos1, bluenoise);

            sceneCol.rgb                = sceneCol.rgb * waterVolumeFogTransmittance;
        }
    #else
        if (sceneDepth0 < sceneDepth1) {
            vec4 tex1                   = stex(colortex1);
            vec4 translucentColor       = stex(colortex3);
            vec3 translucentAlbedo      = sqr(decode3x8(tex1.b));
            sceneCol.rgb                = blendTranslucencies(sceneCol.rgb, translucentColor, translucentAlbedo);
        }
    #endif

    sceneColor         = make_drawbuffer(sceneCol);
}
