#version 450 compatibility

/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:0*/
layout(location = 0) out vec4 sceneColor;


/* ------ inputs from vertex stage ------ */
in vec2 coord;


/* ------ uniforms ------ */
uniform int frameCounter;
uniform int isEyeInWater;

uniform float eyeAltitude;
uniform float far, near;
uniform float frameTimeCounter;
uniform float rainStrength;
uniform float wetness;

uniform ivec2 eyeBrightness;

uniform vec2 pixelSize;
uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 lightVector, lightVectorView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;

uniform sampler2D depthtex0;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;

uniform sampler2D noisetex;


/* ------ includes ------*/
#include "/lib/util/bicubic.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/fragment/labPBR.glsl"
#include "/lib/lighting/warp.glsl"
#include "/lib/lighting/brdf.glsl"


/* ------ functions ------ */
vec3 decode3x8(float a){
    int bf = int(a*65535.);
    return vec3(bf%32, (bf>>5)%64, bf>>11) / vec3(31,63,31);
}


/* ------ spatial upscale for fog ------ */
float depth_lin(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

const float kernelW_3x3[9]  = float[9](
    0.0779, 0.12325, 0.0779,
    0.12325, 0.1954,  0.12225,
    0.0779, 0.12325, 0.0779
);

const ivec2 kernelO_3x3[9]  = ivec2[9](
    ivec2(-1, -1), ivec2(0, -1), ivec2(1, -1),
    ivec2(-1, 0),  ivec2(0, 0),  ivec2(1, 0),
    ivec2(-1, 1),  ivec2(0, 1),  ivec2(1, 1)
);

mat2x3 spatialUpscale(sampler2D tex1, sampler2D tex2, sampler2D depth, const float lod, float sdepth, const vec2 offset) {
    mat2x3 total = mat2x3(0.0);
    float sum   = 0.0;
    ivec2 depth_pos     = ivec2(coord * viewSize);
    ivec2 tex_pos       = ivec2((coord * rcp(lod) + offset) * viewSize);

        sdepth  = depth_lin(sdepth);

    for (int i = 0; i<9; i++) {
        ivec2 d  = kernelO_3x3[i];

        //float blur  = expf(-dot(d, d) * 0.1);

        ivec2 dp    = depth_pos + ivec2(d * lod);
        float cdepth    = texelFetch(depth, dp, 0).x;
        float cdlin     = depth_lin(cdepth);
        float weight    = expf(-max(distance(sdepth, cdlin) * far - 0.15, 0.0) * 0.5);
        ivec2 tp    = tex_pos + d;
        total[0]   += texelFetch(tex1, tp, 0).rgb * weight;
        total[1]   += texelFetch(tex2, tp, 0).rgb * weight;
        sum        += weight;
    }

    return total * rcp(max(sum, 1e-20));
}


/* ------ reflections ------ */
bool raytraceIntersection(in vec3 pos, in vec3 dir, out vec3 hitScreenPosition, out vec3 hitViewPosition, const uint maxSteps, const uint maxRefs, const vec3 param) {
    //This is a WiP. It is the almost same as the one in Ozius iirc, but that will change.
    /*const uint maxSteps = 32u;
    const float maxRefs = 4;
    const float stepSize = 0.5;
    const float stepScale = 1.6;
    const float refScale = 0.1;*/
    const float stepSize = param.x;
    const float stepScale = param.y;
    const float refScale = param.z;

    vec3 increment = dir * stepSize;

    hitViewPosition = pos;

    uint refinements = 0u;
    for(uint i = 0u; i < maxSteps; ++i) {
        hitViewPosition += increment;
        hitScreenPosition = viewSpaceToScreenSpace(hitViewPosition, gbufferProjection);

        if (any(greaterThan(abs(hitScreenPosition - 0.5), vec3(0.5)))) return false;

        float depth = texelFetch(depthtex0, ivec2(hitScreenPosition.xy * viewSize), 0).r;
        float diff = hitViewPosition.z - screenSpaceToViewSpace(depth, gbufferProjectionInverse);

        if(diff <= 0.0) {
            if(++refinements < maxRefs) {
                hitViewPosition -= increment;
                increment *= refScale;

                continue;
            }

            if (any(greaterThan(abs(hitScreenPosition - 0.5), vec3(0.5))) || length(increment) * 10 < -diff || depth == 1.0) return false;

			return true;
        }

        increment *= stepScale;
    }

    return false;
}

bool raytraceIntersection(in vec3 pos, in vec3 dir, out vec3 hitScreenPosition, out vec3 hitViewPosition) {
    return raytraceIntersection(pos, dir, hitScreenPosition, hitViewPosition, 32u, 4u, vec3(0.5, 1.6, 0.1));
}

mat3 rotationMatrix(vec3 x, vec3 y) {
	float cosine = dot(x, y);
	vec3 axis = cross(y, x);

	float tmp = 1.0 / dot(axis, axis);
	      tmp = tmp - tmp * cosine;
	vec3 tmpv = axis * tmp;

	return mat3(
		axis.x * tmpv.x + cosine, axis.x * tmpv.y - axis.z, axis.x * tmpv.z + axis.y,
		axis.y * tmpv.x + axis.z, axis.y * tmpv.y + cosine, axis.y * tmpv.z - axis.x,
		axis.z * tmpv.x - axis.y, axis.z * tmpv.y + axis.x, axis.z * tmpv.z + cosine
	);
}

vec3 ggxFacetDist(vec3 viewDir, float roughness, vec2 xy) {
	/*
    GGX VNDF sampling
	http://www.jcgt.org/published/0007/04/01/
    */

    viewDir     = normalize(vec3(roughness * viewDir.xy, viewDir.z));

    float clsq  = dot(viewDir.xy, viewDir.xy);
    vec3 T1     = vec3(clsq > 0.0 ? vec2(-viewDir.y, viewDir.x) * inversesqrt(clsq) : vec2(1.0, 0.0), 0.0);
    vec3 T2     = vec3(-T1.y * viewDir.z, viewDir.z * T1.x, viewDir.x * T1.y - T1.x * viewDir.y);

	float r     = sqrt(xy.x);
	float phi   = tau * xy.y;
	float t1    = r * cos(phi);
	float a     = saturate(1.0 - t1 * t1);
	float t2    = mix(sqrt(a), r * sin(phi), 0.5 + 0.5 * viewDir.z);

	vec3 normal_h = (t1 * T1) + (t2 * T2) + sqrt(saturate(a - t2 * t2)) * viewDir;

	return normalize(vec3(roughness * normal_h.xy, normal_h.z));
}

mat2x3 roughReflection(vec3 viewNormal, vec3 viewPos, pbr_data pbr, float dither, float skyfade, out float alpha) {
    if (pbr.roughnessSquared > 0.9) return mat2x3(0.0);

    float rfade     = 1.0 - linStep(pbr.roughnessSquared, 0.6, 0.9);
    
    uint steps      = 2 + uint(min(pbr.roughnessSquared * 3.0, 2.0));
    float rstep     = 1.0 / float(steps);

    mat3 rot        = rotationMatrix(vec3(0, 0, 1), viewNormal);
    vec3 tangentV   = normalize(viewPos) * rot;

    vec3 reflection = vec3(0.0);
    vec3 fresnel    = vec3(0.0);
        alpha       = 0.0;

    for (uint i = 0; i < steps; ++i) {
        vec2 xy     = vec2(fract((i + dither) * sqr(32.0) * phi), (i + dither) * rstep);
        vec3 roughNrm = rot * ggxFacetDist(-tangentV, pbr.roughnessSquared, xy);

        vec3 reflectedPos   = reflect(viewPos, roughNrm);

        fresnel    += get_fresnel(pbr.eta, pbr.etak, dot(normalize(reflectedPos), roughNrm), (pbr.is_metal && !pbr.generic_metal));

        vec3 screenHitPosition;
        vec3 viewHitPosition;

        bool hit = false;

        #ifdef ssr_enabled
        if (pbr.roughnessSquared >= 0.75) {
            hit = raytraceIntersection(viewPos, reflectedPos, screenHitPosition, viewHitPosition, 12u, 4u, vec3(0.5, 1.6, 0.1));
        } else if (pbr.roughnessSquared >= 0.5) {
            hit = raytraceIntersection(viewPos, reflectedPos, screenHitPosition, viewHitPosition, 16u, 4u, vec3(0.5, 1.6, 0.1));
        } else if (pbr.roughnessSquared >= 0.25) {
            hit = raytraceIntersection(viewPos, reflectedPos, screenHitPosition, viewHitPosition, 20u, 4u, vec3(0.5, 1.6, 0.1));
        } else if (pbr.roughnessSquared >= 0.125) {
            hit = raytraceIntersection(viewPos, reflectedPos, screenHitPosition, viewHitPosition, 26u, 4u, vec3(0.5, 1.6, 0.1));
        } else {
            hit = raytraceIntersection(viewPos, reflectedPos, screenHitPosition, viewHitPosition, 32u, 4u, vec3(0.5, 1.6, 0.1));
        }
        #endif

        if (hit) {
            reflection += texture(colortex0, screenHitPosition.xy).rgb;
            alpha      += 1.0;
        }
    }

    reflection *= rstep;
    fresnel    *= rstep;
    alpha      *= rstep;

    return mat2x3(reflection, fresnel);
}

mat2x3 glossyReflection(vec3 viewNormal, vec3 viewPos, pbr_data pbr, float skyfade, out float alpha) {
    vec3 reflectedPos   = reflect(viewPos, viewNormal);

    vec3 fresnel = get_fresnel(pbr.eta, pbr.etak, dot(normalize(reflectedPos), viewNormal), (pbr.is_metal && !pbr.generic_metal));

    vec3 screenHitPosition;
    vec3 viewHitPosition;

    #ifdef ssr_enabled
    bool hit = raytraceIntersection(viewPos, reflectedPos, screenHitPosition, viewHitPosition);
    #else
    const bool hit = false;
    #endif

    vec3 reflection = vec3(0.0);
    alpha   = 0.0;

    if (hit) {
        reflection += texture(colortex0, screenHitPosition.xy).rgb;
        alpha      += 1.0;
    }

    return mat2x3(reflection, fresnel);
}

void main() {
    vec4 sceneCol   = stex(colortex0);
    float sceneDepth = stex(depthtex0).r;

    vec3 viewPos    = screenSpaceToViewSpace(vec3(coord, sceneDepth), gbufferProjectionInverse, true);
    vec3 scenePos   = viewSpaceToSceneSpace(viewPos, gbufferModelViewInverse);


    /* --- reflections --- */
    if (landMask(sceneDepth)) {
        vec4 tex1           = stex(colortex1);
        vec4 tex2           = stex(colortex2);
        pbr_data pbr        = decode_lab(vec4(decode2x8(tex2.z), decode2x8(tex1.w)));

        if (pbr.f0 > 0.0) {
            vec3 sceneNormal    = decodeNormal(tex1.xy);
            vec2 sceneLmap      = decode2x8(tex1.z);
            vec3 viewNormal     = mat3(gbufferModelView) * sceneNormal;

            mat2x3 reflection   = mat2x3(0.0);
            float alpha         = 0.0;

            if (pbr.roughnessSquared < 2e-3) {
                reflection      = glossyReflection(viewNormal, viewPos, pbr, 0.0, alpha);
            } else {
                reflection      = roughReflection(viewNormal, viewPos, pbr, dither_bluenoise(), 0.0, alpha);
            }

            if (pbr.is_metal) sceneCol.rgb *= 1.0 - alpha;

            if (pbr.generic_metal) {
                vec3 albedo     = decode3x8(tex2.a);
                reflection[0]  *= albedo;
            }

            sceneCol.rgb        = mix(sceneCol.rgb, reflection[0], reflection[1]);
        }
    }

    sceneColor      = make_drawbuffer(sceneCol);
}