#version 450 compatibility

/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"

const int shadowMapResolution   = 2560;     //[512 1024 1536 2048 2560 3072 3584 4096 6144 8192 16384]
const float shadowDistance      = 128.0; //we'll use robobos infinite shadowdistance warping anyways


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:023*/
layout(location = 0) out vec4 scene_color;
layout(location = 1) out vec4 auxiliary_data;
layout(location = 2) out vec4 translucency_data;


/* ------ inputs from vertex stage ------ */
in vec2 coord;

flat in vec3 blockLightColor;


/* ------ uniforms ------ */
uniform int frameCounter;

uniform float far, near;
uniform float nightVision;

uniform vec2 pixelSize;
uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;

uniform sampler2D depthtex0;

uniform sampler2D noisetex;


/* ------ includes ------ */
#include "/lib/util/encoders.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/fragment/dither.glsl"


/* ------ lighting calculation ------ */
vec3 get_blocklight(vec3 color, float lightmap) {
    lightmap    = pow5(linStep(lightmap, rcp(24.0), 1.0));

    return lightmap * color;
}


/* ------ labPBR ------ */
#include "/lib/fragment/labPBR.glsl"


/* ------ upscaling for ao and gi ------ */
float depth_lin(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

vec4 bilateral_upscale(sampler2D tex, sampler2D depth, const int lod, float sdepth) {
    vec4 total  = vec4(0.0);
    float sum   = 0.0;
    ivec2 depth_pos     = ivec2(coord * viewSize);
    ivec2 tex_pos       = ivec2(coord * viewSize * rcp(float(lod)));

        sdepth  = depth_lin(sdepth);

    const int size  = 1;

    for (int i = -size; i<=size; i++) {
        for (int j = -size; j<=size; j++) {
            ivec2 dp    = depth_pos + ivec2(i, j) * lod;
            float cdepth    = texelFetch(depth, dp, 0).x;
            float cdlin     = depth_lin(cdepth);
            float weight    = landMask(cdepth) ? expf(-max(distance(sdepth, cdlin) * far - 0.05, 0.0) * 1.0) : 1e-16;
            ivec2 tp    = tex_pos + ivec2(i, j);
            total      += texelFetch(tex, tex_pos, 0) * weight;
            sum        += weight;
        }
    }

    return total * rcp(max(sum, 1e-20));
}


/* ------ encoding for generic metals ------ */
float bayer2e(vec2 a) {
    a = floor(a);
    return fract( dot(a, vec2(.5, a.y * .75)) );
}

#define bayer4e(a) (bayer2e( .5*(a))*.25+bayer2e(a))

#define m vec3(31,63,31)
float encode3x8(vec3 a) {
    float dither = bayer4e(gl_FragCoord.xy);
    a += (dither-.5) / m;
    a = saturate(a);
    ivec3 b = ivec3(a*m);
    return float( b.r|(b.g<<5)|(b.b<<11) ) / 65535.;
}
#undef m

void main() {
    vec4 sceneCol   = stex(colortex0);
    float sceneDepth = stex(depthtex0).r;

    vec4 tex2       = stex(colortex2);

    vec3 albedo     = sceneCol.rgb;

    if (landMask(sceneDepth)) {
        vec4 tex1       = stex(colortex1);
        vec3 sceneNormal = decodeNormal(tex1.xy);
        vec2 sceneLmap  = decode2x8(tex1.z);

        vec3 viewPos    = screenSpaceToViewSpace(vec3(coord, sceneDepth), gbufferProjectionInverse, true);
        vec3 scenePos   = viewSpaceToSceneSpace(viewPos, gbufferModelViewInverse);

        pbr_data pbr    = decode_lab(vec4(decode2x8(tex2.z), decode2x8(tex1.w)));

        #ifdef ambient_occlusion_enabled
            vec4 tex3   = bilateral_upscale(colortex3, depthtex0, 2, sceneDepth);
            float ao    = tex3.a;
        #else
            float ao    = sceneCol.a;
        #endif

        float bluenoise     = dither_bluenoise();

        vec3 albedo     = sceneCol.rgb;

        vec3 ambient_min    = vec3(0.3, 0.6, 1.0) * ((0.00026 + ceil(nightVision) * 0.005));

        vec3 block_light    =  get_blocklight(blockLightColor, sceneLmap.x);

        vec3 total_light    = (ambient_min + block_light) * albedo * ao;

        sceneCol.rgb        = total_light;
    }

    scene_color     = make_drawbuffer(sceneCol.rgb);
    translucency_data = vec4(0.0);

    auxiliary_data  = tex2;
    auxiliary_data.a = encode3x8(albedo);
}