#version 450 compatibility

/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:034*/
layout(location = 0) out vec4 sceneColor;
layout(location = 1) out vec4 fogScattering;
layout(location = 2) out vec4 fogTransmittance;


/* ------ inputs from vertex stage ------ */
in vec2 coord;

flat in float lightFlip;

flat in mat3x3 lightColor;


/* ------ uniforms ------ */
uniform int frameCounter;
uniform int isEyeInWater;

uniform float eyeAltitude;
uniform float frameTimeCounter;

uniform ivec2 eyeBrightness;

uniform vec2 pixelSize;
uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 cameraPosition;
uniform vec3 fogData;
uniform vec3 lightVector, lightVectorView;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex6;

uniform sampler2D depthtex0;
uniform sampler2D depthtex1;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;


/* ------ includes ------*/
#include "/lib/util/srgb.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/lighting/warp.glsl"
#define PD_3DNOISE
#include "/lib/fragment/noise.glsl"
#include "/lib/atmosphere/constants.glsl"
#include "/lib/atmosphere/phase.glsl"



/* ------ functions ------ */
vec3 decode3x8(float a){
    int bf = int(a*65535.);
    return vec3(bf%32, (bf>>5)%64, bf>>11) / vec3(31,63,31);
}


/* ------ shadow position transforms ------ */
vec3 sceneToShadowSpace(vec3 scenePos, const float bias, out float cd) {
    float a     = length(scenePos);
    vec3 pos    = scenePos + bias * lightVector;
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD3(shadowProjection, pos);
        pos.z  *= shadowFarPlaneMult;
        pos.z  -= 0.0012*(saturate(a/256.0));
        float warp = 1.0;
        pos.xy  = warp_shadowmap(pos.xy, warp);

    vec3 temp   = pos;
        temp.xy *= warp;
        temp    = projMAD3(shadowProjectionInverse, temp);
        cd      = -temp.z;

    return pos * 0.5 + 0.5;
}
vec2 cloudShadowSpace(vec3 scenePos) {
    vec3 pos    = mat3(shadowModelView) * scenePos;
        pos    *= rcp(vcloud_shadow_clip);
        pos.xy  = pos.xy * 0.5 + 0.5;

    vec2 mcoeff = viewSize * rcp(1024.0);

    return pos.xy * rcp(mcoeff);
}


/* ------ shadow depth comparision ------*/
float get_shadowtex(sampler2D tex, vec2 coord) {
    return texture(tex, coord).x;
}
float get_shadow(float depth, float cd) {
    return 1.0-saturate((cd-depth*256.0-0.0085)*25.0);
}
float get_shadow(sampler2D tex, vec2 coord, float cd) {
    return get_shadow(get_shadowtex(tex, coord).x, cd);
}
vec3 get_shadowcol(sampler2D tex, vec2 coord) {
    vec4 x  = texture(tex, coord);
    return mix(vec3(1.0), x.rgb, x.a);
}


/* ------ volumetric fog ------ */
const float fogAirDensity   = 2e5;
const float fogMaxalt   = vfog_alt + 90.0;
vec3 fogWind    = vec3(frameTimeCounter, frameTimeCounter * 0.1, frameTimeCounter * 0.03) * 0.28;

const vec2 fogDensity = 6e-4 * vec2(vfog_rayl_density, vfog_mie_density);
const float fogMistDensity = 1.0e-2 * vfog_mist_density;

vec3 atmosFogDensity(vec3 pos) {
    float fade  = expf(-max(pos.y - vfog_alt, 0.0) * pi * 3e-3);
        fade   *= 1.0 - sstep(pos.y, fogMaxalt - 20.0, fogMaxalt);

    float mist  = 0.0;

    if (fogMistDensity > 1e-6) {
        float mistFade  = expf(-max(pos.y - vfog_alt, 0.0) * tau * 8e-3);
            mistFade   *= 1.0 - sstep(pos.y, vfog_alt + 25.0, vfog_alt + 45.0);
        
        if (mistFade > 1e-6) {
            pos    += fogWind;
            pos    *= 0.0036;
            vec3 noise  = noise_3d(pos);
                mist   += (noise.z * 0.95) + (noise.y * 0.1) - ((0.32 * (1.0 - mistFade)) + 0.34);
                mist    = max(mist * mistFade, 0.0);
        }
    }

    #ifdef vfog_dynamic_density
        return vec3(fogDensity * fade, fogMistDensity * mist) * fogData;
    #else
        return vec3(fogDensity * fade, fogMistDensity * mist);
    #endif
}

float atmosFogPhase(float cosTheta) {
    float a = cornettShanksMiePhase(cosTheta, mieG);
    float b = cornettShanksMiePhase(cosTheta, -0.45);

    return mix(a, b, 0.18) + 0.022;
}

mat2x3 compute_atmosFog(vec3 scenePos, vec3 worldDir, bool isSky, float vDotL, float dither, mat2x3 lightColors) {    //fog in front of and behind translucents are separated to have less branching going on
    float ts        = length(worldDir*((fogMaxalt-eyeAltitude)/worldDir.y));
    float bs        = length(worldDir*((-64.0-eyeAltitude)/worldDir.y));

    float edgeDist  = worldDir.y > 0.0 ? ts : bs;

    float edist     = isSky ? min(edgeDist, vfog_maxdist) : length(scenePos);
    float sdist     = eyeAltitude > fogMaxalt ? ts : 1.0;

    vec3 spos       = eyeAltitude > fogMaxalt ? (worldDir * sdist) + gbufferModelViewInverse[3].xyz : gbufferModelViewInverse[3].xyz;
    vec3 epos       = isSky ? (worldDir * edist) : scenePos;

    float bstep      = length((epos-spos));
    float stepcoeff  = saturate(bstep/256.0);
    uint steps       = uint((12 + stepcoeff * 12) * vfog_quality);

    vec3 rstep      = (epos-spos) / float(steps);
    vec3 rpos       = (rstep * dither) + spos;
    float rlength   = length(rstep);

    mat2x3 scattering = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    vec3 phase      = vec3(phaseFunction(vDotL), atmosFogPhase(vDotL));
    float phaseIso  = 0.25 * rcp(pi);

    for (uint i = 0; i < steps; i++, rpos += rstep) {
        if (v3avg(transmittance) < 5e-3) break;
        if ((rpos.y + cameraPosition.y) > fogMaxalt) continue;

        vec3 density    = atmosFogDensity(rpos + cameraPosition);

        vec3 od = rlength * density * fogAirDensity;
        vec3 stepOpticalDepth = fogExtinctionCoefficients * od;

        vec3 stepTransmittance = saturate(expf(-stepOpticalDepth));
        vec3 stepTransmittedFraction = saturate((stepTransmittance - 1.0) / -max(stepOpticalDepth, 1e-16));

        vec3 visibleScattering = stepTransmittedFraction;
            visibleScattering *= transmittance;

        vec3 stepScatteringView = fogScatteringCoefficients * (od * phase);
        stepScatteringView *= visibleScattering;
        vec3 stepScatteringViewSky = fogScatteringCoefficients * (od * phaseIso);
        stepScatteringViewSky *= visibleScattering;

        float cd    = 0.0;
        vec3 shadowPos  = sceneToShadowSpace(rpos, 0.08, cd);

        vec3 directLight    = vec3(1.0);

        if (cd < 256.0) {
            float shadow0       = get_shadow(shadowtex0, shadowPos.xy, cd);

            if (shadow0 < 1.0) {
                float shadow1   = get_shadow(shadowtex1, shadowPos.xy, cd);

                directLight    *= shadow1;

                if (distance(shadow0, shadow1) > 0.1) {
                    directLight    *= get_shadowcol(shadowcolor0, shadowPos.xy);
                }
            }
        }

        #ifdef cloud_shadows
            directLight    *= texture(colortex6, cloudShadowSpace(rpos)).a * 0.9 + 0.1;
        #endif

        scattering[0]  += stepScatteringView * transmittance * directLight;
        scattering[1]  += stepScatteringViewSky * transmittance;
        transmittance  *= stepTransmittance;
    }

    vec3 color      = scattering[0] * lightColors[0] + scattering[1] * lightColors[1];

    return mat2x3(color, transmittance);
}


/* ------ water volume ------ */
float fournierForandPhase(float phi, float mu, float n) {
	float v = (3.0 - mu) / 2.0;
	float delta = (4.0 / (3.0 * sqr(n - 1))) * sqr(sin(phi / 2.0));
	float delta180 = (4.0 / (3.0 * sqr(n - 1))) * sqr(sin(pi / 2.0));

	float p1 = 1.0 / (4.0 * pi * sqr(1.0 - delta) * pow(delta, v));
	float p2 = v * (1.0 - delta) - (1.0 - pow(delta, v)) + (delta * (1.0 - pow(delta, v)) - v * (1.0 - delta)) * pow(sin(phi / 2.0), -2.0);
	float p3 = ((1.0 - pow(delta180, v)) / (16.0 * pi * (delta180 - 1.0) * pow(delta180, v))) * (3.0 * sqr(cos(phi)) - 1.0);
	return p1 * p2 + p3;
}

mat2x3 compute_waterVolume(vec3 scenePos0, vec3 scenePos1, float dither, mat2x3 lightColors) {
    bool isUnderwater   = isEyeInWater == 1;

    vec3 epos   = isUnderwater ? scenePos0 : scenePos1;
    vec3 spos   = isUnderwater ? gbufferModelViewInverse[3].xyz : scenePos0;

    float bstep      = length((epos-spos));
    float stepcoeff  = saturate(bstep * rcp(64.0));
    uint steps       = 4 + uint(stepcoeff * 6);

    vec3 rstep      = (epos-spos)/steps;
    vec3 rpos       = rstep*dither+spos;
    float rlength   = length(rstep);

    mat2x3 scattering = mat2x3(0.0);
    vec3 transmittance = vec3(1.0);

    float phase     = fournierForandPhase(acosFast(-dot(normalize(lightVector), -normalize(epos - gbufferModelViewInverse[3].xyz))), 3.483, 1.08);
    float phaseIso  = 0.25 * rcp(pi);
        phase       = phaseIso * 0.7 + phase * 0.3;

    const vec3 absorbCoeff = vec3(2e-1, 4e-2, 1e-2);
    const vec3 scatterCoeff = vec3(1.5e-2);
    const vec3 attenCoeff = (absorbCoeff + scatterCoeff) * 4.0;

    vec3 skylight   = lightColors[1] * phaseIso;

    for (uint i = 0; i < steps; i++, rpos += rstep) {
        if (max(transmittance.x, max(transmittance.y, transmittance.z)) < 0.01) break;

        float od    = rlength;

        float cd    = 0.0;
        vec3 shadowPos  = sceneToShadowSpace(rpos, 0.08, cd);

        float shadow    = 1.0;

        if (cd < 256.0) {
            shadow      = get_shadow(shadowtex1, shadowPos.xy, cd);
        }

        #ifdef cloud_shadows
            shadow     *= texture(colortex6, cloudShadowSpace(rpos)).a * 0.9 + 0.1;
        #endif

        vec3 uod = abs(rpos.y - spos.y) * attenCoeff;
        vec3 st = expf(-uod * 1.5);

        scattering[0]  += (shadow) * transmittance;
        scattering[1]  += (skylight * st) * transmittance;
        transmittance  *= expf(-attenCoeff * od);
    }

    scattering[0]  *= scatterCoeff;
    scattering[0]  *= (1.0 - expf(-attenCoeff * rlength)) * rcp(attenCoeff);
    scattering[1]  *= scatterCoeff;
    scattering[1]  *= (1.0 - expf(-attenCoeff * rlength)) * rcp(attenCoeff);
    vec3 color      = (scattering[0] * lightColors[0] * phase) + (scattering[1] * skylight);

    return mat2x3(color, transmittance);
}

vec3 blendTranslucencies(vec3 sceneCol, vec4 translucents, vec3 albedo) {
    vec3 color  = sceneCol;
        color  *= mix(vec3(1.0), albedo, translucents.a);
        color   = mix(color, translucents.rgb, translucents.a);

    return color;
}

void main() {
    vec4 sceneCol   = stex(colortex0);
    float sceneDepth0 = stex(depthtex0).r;
    float sceneDepth1 = stex(depthtex1).r;

    vec4 tex2       = stex(colortex2);
    int matID       = decodeMatID16(tex2.x);

    float caveFix   = linStep(eyeBrightness.y/240.0, 0.1, 0.9);

    float bluenoise = dither_bluenoise();

    mat2x3 fogLightColors   = mat2x3(lightColor[0] * lightFlip, lightColor[1] * caveFix);

    vec3 viewPos0    = screenSpaceToViewSpace(vec3(coord, sceneDepth0), gbufferProjectionInverse, true);
    vec3 scenePos0   = viewSpaceToSceneSpace(viewPos0, gbufferModelViewInverse);

    #ifdef vfog_enabled
        vec3 viewPos1    = screenSpaceToViewSpace(vec3(coord, sceneDepth1), gbufferProjectionInverse, true);
        vec3 scenePos1   = viewSpaceToSceneSpace(viewPos1, gbufferModelViewInverse);
        float vDotL     = dot(normalize(viewPos1), lightVectorView);

        mat2x3 volumeFog;

        if (sceneDepth0 < sceneDepth1) {
            vec4 tex1        = stex(colortex1);
            vec4 translucentColor   = stex(colortex3);
            vec3 translucentAlbedo  = sqr(decode3x8(tex1.b));
            if (matID == 101 && isEyeInWater == 0) volumeFog = compute_waterVolume(scenePos0, scenePos1, bluenoise, fogLightColors);
            else volumeFog  = mat2x3(vec3(0.0), vec3(1.0));

            sceneCol.rgb    = sceneCol.rgb * volumeFog[1] + volumeFog[0];

            sceneCol.rgb    = blendTranslucencies(sceneCol.rgb, translucentColor, translucentAlbedo);
        } 
        
        if (isEyeInWater == 1) {
            volumeFog = compute_waterVolume(scenePos0, scenePos1, bluenoise, fogLightColors);

            sceneCol.rgb    = sceneCol.rgb * volumeFog[1] + volumeFog[0];
        }
    #else
        if (sceneDepth0 < sceneDepth1) {
            vec4 tex1        = stex(colortex1);
            vec4 translucentColor   = stex(colortex3);
            vec3 translucentAlbedo  = sqr(decode3x8(tex1.b));
            sceneCol.rgb    = blendTranslucencies(sceneCol.rgb, translucentColor, translucentAlbedo);
        }
    #endif

    fogScattering       = vec4(0.0, 0.0, 0.0, 1.0);
    fogTransmittance    = stex(colortex4);

    #ifdef vfog_enabled
    const float fLOD    = 2.0;

    vec2 scaledCoord    = (coord - vec2(0.5, 0.5)) * fLOD;

    if (clamp(scaledCoord, -pixelSize * fLOD, 1.0 + pixelSize * fLOD) == scaledCoord) {
        vec2 coord      = saturate(scaledCoord);

        float sceneDepth0 = texelFetch(depthtex0, ivec2(coord * viewSize), 0).r;
        vec3 viewPos0    = screenSpaceToViewSpace(vec3(coord, sceneDepth0), gbufferProjectionInverse, true);
        vec3 scenePos0   = viewSpaceToSceneSpace(viewPos0, gbufferModelViewInverse);

        mat2x3 volumeFog;

        float vDotL     = dot(normalize(viewPos0), lightVectorView);

        volumeFog       = compute_atmosFog(scenePos0, normalize(scenePos0), !landMask(sceneDepth0), vDotL, bluenoise, fogLightColors);

        fogScattering.rgb       = volumeFog[0];
        fogTransmittance.rgb    = volumeFog[1];
    }
    #endif

    sceneColor          = make_drawbuffer(sceneCol);
    fogScattering       = clamp_drawbuffer(fogScattering);
    fogTransmittance    = clamp_drawbuffer(fogTransmittance);
}