#version 450 compatibility

/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"

const int shadowMapResolution   = 2560;     //[512 1024 1536 2048 2560 3072 3584 4096 6144 8192 16384]
const float shadowDistance      = 128.0; //we'll use robobos infinite shadowdistance warping anyways


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:023*/
layout(location = 0) out vec4 scene_color;
layout(location = 1) out vec4 auxiliary_data;
layout(location = 2) out vec4 translucency_data;


/* ------ inputs from vertex stage ------ */
in vec2 coord;

flat in float lightFlip;

flat in mat3x3 lightColor;

#if ibl_mode == 1
    flat in mat4x3 IBL_light;
#endif


/* ------ uniforms ------ */
uniform int frameCounter;

uniform float far, near;
uniform float nightVision;
uniform float frameTimeCounter;

uniform vec2 pixelSize;
uniform vec2 viewSize;
uniform vec2 taaOffset;

uniform vec3 lightVector, lightVectorView;
uniform vec3 sunVector;
uniform vec3 moonVector;
uniform vec3 upVector;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 shadowModelView, shadowModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
#if ibl_mode >=2
    uniform sampler2D colortex4;
#endif
#ifdef cloud_shadows
    uniform sampler2D colortex6;
#endif

uniform sampler2D depthtex0;

uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;

uniform sampler2D noisetex;


/* ------ includes ------ */
#include "/lib/util/srgb.glsl"
#include "/lib/util/encoders.glsl"
#include "/lib/util/transforms.glsl"
#include "/lib/util/bicubic.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/lighting/diffuse.glsl"
#include "/lib/lighting/warp.glsl"


/* ------ shadow position transforms ------ */
vec3 sceneToShadowSpace(vec3 scenePos, const float bias, out float warp, out float cd) {
    float a     = length(scenePos);
    vec3 pos    = scenePos + bias * lightVector;
        pos     = transMAD(shadowModelView, pos);
        pos     = projMAD3(shadowProjection, pos);
        pos.z  *= shadowFarPlaneMult;
        pos.z  -= 0.0012*(saturate(a/256.0));
        
        pos.xy  = warp_shadowmap(pos.xy, warp);

    vec3 temp   = pos;
        temp.xy *= warp;
        temp    = projMAD3(shadowProjectionInverse, temp);
        cd      = -temp.z;

    return pos * 0.5 + 0.5;
}

vec3 sceneToShadowView(vec3 pos, const float bias) {
        pos     = pos + bias * lightVector;
        pos     = transMAD(shadowModelView, pos);
    return pos;
}
vec3 shadowViewToProjection(vec3 pos, out float warp, out float cd) {
        pos     = projMAD3(shadowProjection, pos);
        pos.z  *= shadowFarPlaneMult;
        
        pos.xy  = warp_shadowmap(pos.xy, warp);

    vec3 temp   = pos;
        temp.xy *= warp;
        temp    = projMAD3(shadowProjectionInverse, temp);
        cd      = -temp.z;

    return pos * 0.5 + 0.5;
}

vec2 cloudShadowSpace(vec3 scenePos) {
    vec3 pos    = mat3(shadowModelView) * scenePos;
        pos    *= rcp(vcloud_shadow_clip);
        pos.xy  = pos.xy * 0.5 + 0.5;

    vec2 mcoeff = viewSize * rcp(1024.0);

    return pos.xy * rcp(mcoeff);
}


/* ------ shadow depth comparision ------*/
float get_shadowtex(sampler2D tex, vec2 coord) {
    return texture(tex, coord).x;
}
float get_shadowtex_cubic(sampler2D tex, vec2 coord) {
    return textureBicubic(tex, coord).x;
}
float get_shadow(float depth, float cd) {
    return 1.0-saturate((cd-depth*256.0 - 0.0085 * shadowBiasMult)*25.0);
}
float get_shadow(sampler2D tex, vec2 coord, float cd) {
    return get_shadow(get_shadowtex(tex, coord).x, cd);
}
vec3 get_shadowcol(sampler2D tex, vec2 coord) {
    vec4 x  = textureBicubic(tex, coord);
    return mix(vec3(1.0), x.rgb, x.a);
}


/* ------ lighting calculation ------ */
#include "/lib/lighting/filter.glsl"

vec4 project_directLight(bool diffuseLit, vec3 scenePos, float cloudShadow) {
    const float bias    = 0.03 * (2048.0 / shadowMapResolution) * shadowBiasMult;
    //const float farplane    = 256.0 / 0.2;  //this is the shadowmap far plane

    if (diffuseLit && cloudShadow > 0.0) {
        float warp          = 1.0;
        float cd            = 0.0;
        vec3 shadowPos      = sceneToShadowSpace(scenePos, bias, warp, cd);

        //cd -= (length(scenePos) / 256.0) * 0.01;

        if (cd > 256.0) return vec4(1.0);

        #ifdef shadow_vps_enabled
            float sigma     = get_vps_sigma(shadowtex1, shadowPos, cd, warp);
        #else
            float sigma     = 0.0004 * rcp(warp);
        #endif

        vec4 shadow     = get_shadow30(shadowPos.xy, cd, sigma);
            shadow.rgb  = SRGBtoACEScg(shadow.rgb);

        return shadow;
    }

    return vec4(1.0);
}

vec3 get_blocklight(vec3 color, float lightmap) {
    lightmap    = pow5(linStep(lightmap, rcp(24.0), 1.0));

    return lightmap * color;
}


/* ------ labPBR ------ */
#include "/lib/fragment/labPBR.glsl"


/* ------ image based lighting ------ */
#if ibl_mode == 1
    vec3 ibl_approximated(vec3 scenenormal, vec3 skylight, mat4x3 IBL_light, float lmap) {
        float weight_0  = max(0.0, dot(scenenormal, vec3(1.0, 0.0, 0.0)));
        float weight_1  = max(0.0, dot(scenenormal, vec3(0.0, 0.0, 1.0)));
        float weight_2  = max(0.0, dot(scenenormal, vec3(-1.0, 0.0, 0.0)));
        float weight_3  = max(0.0, dot(scenenormal, vec3(0.0, 0.0, -1.0)));
        float weight_4  = max(0.0, dot(scenenormal, vec3(0.0, 1.0, 0.0)));
        float weight_5  = max(0.0, dot(scenenormal, vec3(0.0, -1.0, 0.0)));

        vec3 IBL    = vec3(0.0);
            IBL    += IBL_light[0] * weight_0;
            IBL    += IBL_light[1] * weight_1;
            IBL    += IBL_light[2] * weight_2;
            IBL    += IBL_light[3] * weight_3;
            IBL    += skylight * weight_4;
            IBL    += skylight * weight_5 * 0.5;
            IBL     = IBL * 0.6 + skylight * 0.4;

            IBL     = mix(skylight, IBL, lmap);

        return IBL * rcp(pi);
    }
#elif ibl_mode == 2 || ibl_mode == 3
    #include "/lib/atmosphere/project.glsl"

    vec2 unproject_sphere(vec3 dir) {
        vec2 lonlat     = vec2(atan(-dir.x, dir.z), acos(dir.y));
        return lonlat * vec2(rcp(tau), rcp(pi)) + vec2(0.5, 0.0);
    }
    vec3 get_projected_skycloud(vec3 x) {
        vec2 spherecoord = unproject_sphere(x);
        spherecoord = spherecoord*rcp(SKYREF_LOD) + vec2(0.5, 0.0);

        return texture(colortex4, spherecoord).rgb;
    }

    vec3 get_projected_sky(vec3 x) {
        //return texture(colortex4, projectSky(x)).rgb;
        return get_projected_skycloud(x);
    }

    vec3 imageBasedLighting(vec3 scenenormal, vec3 skylight, vec3 viewdir, vec3 albedo, float lmap, pbr_data pbr) {
        //Looks kind of dark, but is correct afaik.
        vec3 IBL = vec3(0.0);
        const int iterations = ibl_iterations;    //16 is just waaaay to high, like it just butchers performance
        int weight = 0;
        
        for(float i = 0.0; i < pi; i += pi/iterations) {
            for(float j = 0.0; j < tau; j += tau/iterations) {
                vec3 direction = normalize(vec3(sin(i) * cos(j), sin(i) * sin(j), cos(i)));     //A D V A N C E D  M A T H S
                #if ibl_mode == 2
                    vec3 diffuse = lambertianDiffuse(albedo, scenenormal, direction) / pi;
                #elif ibl_mode == 3
                    vec3 diffuse = orenNayar(pbr.roughness, albedo * (1.0 - pbr.f0), viewdir, direction, scenenormal);
                #endif
                IBL += diffuse * get_projected_sky(direction).rgb;
                weight++;
            }
        }
        IBL /= weight;

        IBL = IBL * tau * 2.0;

        IBL = mix(skylight * albedo, IBL, lmap);     //having sky IBL in areas where the sky is mostly occluded doesn't really make sense, this improves that
        
        return IBL;
    }
#endif


/* ------ upscaling for ao and gi ------ */
float depth_lin(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

vec4 bilateral_upscale(sampler2D tex, sampler2D depth, const int lod, float sdepth) {
    vec4 total  = vec4(0.0);
    float sum   = 0.0;
    ivec2 depth_pos     = ivec2(coord * viewSize);
    ivec2 tex_pos       = ivec2(coord * viewSize * rcp(float(lod)));

        sdepth  = depth_lin(sdepth);

    const int size  = 1;

    for (int i = -size; i<=size; i++) {
        for (int j = -size; j<=size; j++) {
            ivec2 dp    = depth_pos + ivec2(i, j) * lod;
            float cdepth    = texelFetch(depth, dp, 0).x;
            float cdlin     = depth_lin(cdepth);
            float weight    = landMask(cdepth) ? expf(-max(distance(sdepth, cdlin) * far - 0.05, 0.0) * 1.0) : 1e-16;
            ivec2 tp    = tex_pos + ivec2(i, j);
            total      += texelFetch(tex, tex_pos, 0) * weight;
            sum        += weight;
        }
    }

    return total * rcp(max(sum, 1e-20));
}

vec3 gi_noise(vec3 gi) {
    vec2 noisecoord = coord * viewSize;
    float anim  = (frameTimeCounter * 8 * 256);

    vec3 noise  = vec3(0.0);
        noise.r = texture(noisetex, floor(noisecoord+anim*1.8)*rcp(256.0)).x;
        noise.g = texture(noisetex, floor(noisecoord+vec2(-anim, anim)*1.2)*rcp(256.0)).x;
        noise.b = texture(noisetex, floor(noisecoord+vec2(anim, -anim)*1.4)*rcp(256.0)).x;

    return mix(gi, max(gi + v3avg(gi) * (noise - 0.5) * 1.5, vec3(0.0)), 1.0);
}


/* ------ encoding for generic metals ------ */

float bayer2e(vec2 a) {
    a = floor(a);
    return fract( dot(a, vec2(.5, a.y * .75)) );
}

#define bayer4e(a) (bayer2e( .5*(a))*.25+bayer2e(a))

#define m vec3(31,63,31)
float encode3x8(vec3 a) {
    float dither = bayer4e(gl_FragCoord.xy);
    a += (dither-.5) / m;
    a = saturate(a);
    ivec3 b = ivec3(a*m);
    return float( b.r|(b.g<<5)|(b.b<<11) ) / 65535.;
}
#undef m

void main() {
    vec4 sceneCol   = stex(colortex0);
    float sceneDepth = stex(depthtex0).r;

    vec4 tex2       = stex(colortex2);
    int matID       = decodeMatID16(tex2.x);

    vec3 albedo     = sceneCol.rgb;

    if (landMask(sceneDepth)) {
        vec4 tex1       = stex(colortex1);
        vec3 sceneNormal = decodeNormal(tex1.xy);
        vec2 sceneLmap  = decode2x8(tex1.z);

        vec3 viewPos    = screenSpaceToViewSpace(vec3(coord, sceneDepth), gbufferProjectionInverse, true);
        vec3 scenePos   = viewSpaceToSceneSpace(viewPos, gbufferModelViewInverse);

        pbr_data pbr    = decode_lab(vec4(decode2x8(tex2.z), decode2x8(tex1.w)));

        #if (defined ambient_occlusion_enabled || defined rsm_enabled)
            vec4 tex3   = bilateral_upscale(colortex3, depthtex0, 2, sceneDepth);
        #endif

        #ifdef ambient_occlusion_enabled
            float ao    = tex3.a;
        #else
            float ao    = sceneCol.a;
        #endif

        #ifdef rsm_enabled
            vec3 globalIllum    = tex3.rgb;
                globalIllum     = gi_noise(globalIllum);
        #else
            vec3 globalIllum    = vec3(0.0);
        #endif

        float bluenoise     = dither_bluenoise();

        vec3 albedo     = sceneCol.rgb;

        float skyOcclusion  = pow3(sceneLmap.y);

        float nDotL     = max(0.0, dot(sceneNormal, lightVector));
        vec3 diffuse    = orenNayar(pbr.roughnessSquared, albedo * (1.0 - pbr.f0), normalize(-viewPos), lightVectorView, mat3(gbufferModelView) * sceneNormal);

        if (matID == 2 || matID == 3) {
            nDotL       = nDotL * 0.8 + 0.2;
            diffuse     = diffuse * 0.8 + albedo * 0.2;
        }
        if (matID == 4) {
            diffuse     = diffuse * 0.45 + (sqrt(nDotL) * albedo * 0.25);
        }

        #ifdef cloud_shadows
            float cloudShadow   = textureBicubic(colortex6, cloudShadowSpace(scenePos)).a;
        #else
            const float cloudShadow = 1.0;
        #endif

        vec4 direct_data    = project_directLight(nDotL > 0.0, scenePos, cloudShadow);

        #if ibl_mode == 0
            vec3 ibl    = albedo * lightColor[1] * rcp(sqrt(2.0));
        #elif ibl_mode == 1
            vec3 ibl    = ibl_approximated(sceneNormal, lightColor[1], IBL_light, sceneLmap.y);
                ibl     = mix(vec3(v3avg(ibl)) * vec3(1.0, 0.9, 0.85) * 1.1, ibl, cloudShadow * 0.8 + 0.2);
                ibl     = albedo * ibl;
        #elif ibl_mode == 2 || ibl_mode == 3
            vec3 ibl    = imageBasedLighting(sceneNormal, lightColor[1], normalize(-scenePos), albedo, sceneLmap.y, pbr);
        #endif

        vec3 directional    = min(diffuse, direct_data.a * albedo) * cloudShadow;

        float nDotDown      = sqrt(saturate(dot(sceneNormal, vec3(0.0, -1.0, 0.0)) * 0.5 + 0.5));

        vec3 direct_light   = directional * direct_data.rgb;
            globalIllum    += lightColor[0] * rcp(pow6(tau)) * skyOcclusion * nDotDown;
        vec3 bounced_light  = (globalIllum * albedo) * cloudShadow;

        vec3 indirect_light = sqr(skyOcclusion) * ibl * rcp(pidiv2);

        vec3 ambient_min    = vec3(0.3, 0.6, 1.0) * ((0.00026 + ceil(nightVision) * 0.005) * (1.0 - skyOcclusion));

        vec3 block_light    =  get_blocklight(lightColor[2], sceneLmap.x);

        vec3 total_light    = (direct_light + bounced_light) * lightColor[0];
            total_light    += (indirect_light + (ambient_min + block_light) * albedo) * ao;

        sceneCol.rgb        = total_light;

        //sceneCol.rgb        = bounced_light * lightColor[0];
    }

    scene_color     = make_drawbuffer(sceneCol.rgb);
    translucency_data = vec4(0.0);

    auxiliary_data  = tex2;
    auxiliary_data.a = encode3x8(albedo);
}