#version 450 compatibility


/*
====================================================================================================
    __  __    __  ___    ____    ____ 
    \ \/ /   /  |/  /   /  _/   / __ \
     \  /   / /|_/ /    / /    / /_/ /
     / /   / /  / /   _/ /    / _, _/ 
    /_/   /_/  /_/   /___/   /_/ |_|  

    YMIR SHADERS

    Copyright (C) 2020 Jessie, RRe36, Niemand

    All Rights Reserved unless otherwise explicitly stated.


    By downloading this you have agreed to the license and terms of use.
    These can be found inside the included license-file
    or here: https://github.com/ymir-graphics/ymir-releases/blob/master/LICENSE

    Violating these terms may be penalized with actions according to the Digital Millennium
    Copyright Act (DMCA), the Information Society Directive and/or similar laws
    depending on your country.

====================================================================================================
*/



#include "/lib/universal.glsl"


/* ------ framebuffer attachment outputs ------ */
/*DRAWBUFFERS:037*/
layout(location = 0) out vec4 scene_color;
layout(location = 1) out vec4 indirectLight_data;
layout(location = 2) out vec4 temporal_data;


/* ------ inputs from vertex stage ------ */
in vec2 coord;

in vec3 sunVector;


/* ------ uniforms ------ */
uniform int frameCounter;

uniform float eyeAltitude;
uniform float far, near;

uniform vec2 pixelSize, viewSize;
uniform vec2 taaOffset;

uniform vec3 moonVector;
uniform vec3 cameraPosition, previousCameraPosition;

uniform mat4 gbufferModelViewInverse, gbufferModelView;
uniform mat4 gbufferProjectionInverse, gbufferProjection;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;

uniform sampler2D colortex0;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex6;
uniform sampler2D colortex7;

uniform sampler2D depthtex0;

uniform sampler2D noisetex;
uniform sampler3D depthtex2;


/* ------ includes ------ */
#include "/lib/util/encoders.glsl"
#include "/lib/atmosphere/constants.glsl"
#include "/lib/lighting/diffuse.glsl"
#include "/lib/atmosphere/sun_moon.glsl"
#include "/lib/fragment/dither.glsl"
#include "/lib/fragment/lowresCommon.glsl"
#define PD_3DNOISE
#include "/lib/fragment/noise.glsl"
#include "/lib/util/bicubic.glsl"
#include "/lib/util/transforms.glsl"


/* ------ functions ------ */

/* ------ indirect lighting filter ------ */
float depth_lin(float depth) {
    return (2.0*near) / (far+near-depth * (far-near));
}

#define PD_FILTERLOD 2.0
#include "/lib/fragment/waveletFilter.glsl"

vec3 reproject(vec3 sceneSpace) {
    vec3 prevScreenPos = cameraPosition - previousCameraPosition;
    prevScreenPos = sceneSpace + prevScreenPos;
    prevScreenPos = transMAD(gbufferPreviousModelView, prevScreenPos);
    prevScreenPos = transMAD(gbufferPreviousProjection, prevScreenPos) * (0.5 / -prevScreenPos.z) + 0.5;

    return prevScreenPos;
}


void main() {
    vec3 sceneCol   = stex(colortex0).rgb;
    float sceneDepth = stex(depthtex0).x;
    vec4 tex2       = stex(colortex2);
    int matID       = decodeMatID16(tex2.x);

    /* ------ sky projection ------ */
    if (!landMask(sceneDepth)) {
        vec3 viewPos    = screenSpaceToViewSpace(vec3(coord, sceneDepth), gbufferProjectionInverse);
        vec3 scenePos   = viewSpaceToSceneSpace(viewPos, gbufferModelViewInverse);
        vec3 worldDir   = normalize(scenePos);

        vec3 sun    = physicalSun(worldDir);
        sceneCol.rgb   += sun;
    }

    scene_color     = make_drawbuffer(sceneCol);


    /* ------ indirect lighting ------ */
    indirectLight_data  = stex(colortex3);

    #if (defined ambient_occlusion_enabled || defined rsm_enabled)
        vec2 aoCoord    = coord * 2.0;
        
        if (clamp(aoCoord, 0.0 - pixelSize, 1.0 + pixelSize) == aoCoord) {
            aoCoord     = saturate(aoCoord);

            vec4 currentColor       = filter_3x3(colortex3, depthtex0, coord, 1);

            #ifdef rsm_enabled
                vec3 position   = vec3(aoCoord, texelFetch(depthtex0, ivec2(coord * viewSize * 2.0), 0).r);
                    position    = screenSpaceToViewSpace(position, gbufferProjectionInverse);
                    position    = viewSpaceToSceneSpace(position, gbufferModelViewInverse);

                float currentDistance   = length(position);

                vec3 reprojection   = reproject(position);

                bool offscreen      = saturate(reprojection) != reprojection;

                vec4 historyData    = texture(colortex7, reprojection.xy * 0.5);

                float distanceDelta = distance(historyData.a, currentDistance);

                vec3 accumulatedColor   = vec3(0.0);

                float accumulationWeight = (offscreen) ? 1.0 : 1.0 - exp(-max(distanceDelta - 0.1, 0.0) * 2.0);

                if (offscreen) accumulatedColor = currentColor.rgb;
                else accumulatedColor =  mix(historyData.rgb, currentColor.rgb, accumulationWeight * 0.96 + 0.04);

                temporal_data   = clamp_drawbuffer(vec4(accumulatedColor, currentDistance));
                indirectLight_data.rgb  = accumulatedColor;
            #endif
            
            indirectLight_data.a    = currentColor.a;
        }
    #endif

    indirectLight_data  = clamp_drawbuffer(indirectLight_data);
}